

import json
import os
import os.path

from unittest import TestCase
from xml.dom.minidom import parseString

from . import png_bakery
from . import utils

png_assertion = {
    "uid": "123",
    "issuedOn": "2015-04-01",
    "badge": "http://example.org/badge1",
    "recipient": {
        "identity": "test@example.com",
        "hashed": False
    },
    "verify": {
        "type": "hosted",
        "url": "http://example.org/badgeinstance1"
    }
}

svg_assertion = {
    "uid": "abcdef12345",
    "identity": {
        "recipient": "sha256$cbb08ce07dd7345341b9358abb810d29eb790fed",
        "type": "email",
        "hashed": True
    },
    "verify": {
        "type": "hosted",
        "url": "https://example.org/assertion.json"
    },
    "issuedOn": "2013-11-05",
    "badge": "https://example.org/badge.json"
}


class TypeDetectionTests(TestCase):

    def test_detect_svg_type(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'baked_example.svg'
                               )) as image:
            self.assertEqual(utils.check_image_type(image), 'SVG')

    def test_detect_png_type(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'public_domain_heart.png'
                               )) as image:
            self.assertEqual(utils.check_image_type(image), 'PNG')


class PNGBakingTests(TestCase):
    def test_bake_png(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'public_domain_heart.png'
                               )) as image:

            return_file = png_bakery.bake(image, json.dumps(png_assertion))
            self.assertEqual(utils.check_image_type(return_file), 'PNG')
            return_file.seek(0)
            self.assertEqual(png_bakery.unbake(return_file),
                             json.dumps(png_assertion))

    def test_unbake_png(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'baked_heart.png'
                               )) as image:
            assertion = png_bakery.unbake(image)
            self.assertEqual(json.loads(assertion), png_assertion)


class SVGBakingTests(TestCase):
    def test_bake_svg(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'unbaked_example.svg'
                               )) as image:

            return_file = utils.bake(image, json.dumps(svg_assertion))
            self.assertEqual(utils.check_image_type(return_file), 'SVG')
            return_file.seek(0)
            self.assertEqual(utils.unbake(return_file), svg_assertion['verify']['url'])

    def test_unbake_svg(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'baked_example.svg'
                               )) as image:
            verify_url = utils.unbake(image)
            self.assertEqual(verify_url, svg_assertion['verify']['url'])

    def test_svg_roundtrip_hosted(self):
        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'unbaked_example.svg'
                               )) as image:

            svg_string = json.dumps(svg_assertion)
            return_file = utils.bake(image, svg_string)
            self.assertEqual(utils.check_image_type(return_file), 'SVG')
            return_file.seek(0)
            self.assertEqual(utils.unbake(return_file), svg_assertion['verify']['url'])

        return_file.seek(0)
        baked_xml = parseString(return_file.read())
        assertion_node = baked_xml.getElementsByTagName("openbadges:assertion")[0]
        self.assertEqual(assertion_node.attributes['verify'].value, svg_assertion['verify']['url'])

    def test_svg_roundtrip_signed(self):
        signed_string = 'eyJ0eXAiOiJKV1QiLA0KICJhbGciOiJIUzI1NiJ9.eyJpc3MiOiJqb2UiLA0KICJleHAiOjEzMDA4MTkzODAsDQogImh0dHA6Ly9leGFtcGxlLmNvbS9pc19yb290Ijp0cnVlfQ.dBjftJeZ4CVP-mB92K27uhbUJU1p1r_wW1gFWFOEjXk'

        with open(os.path.join(os.path.dirname(__file__),
                               'testfiles', 'unbaked_example.svg'
                               )) as image:

            return_file = utils.bake(image, signed_string)

        return_file.seek(0)
        self.assertEqual(utils.unbake(return_file), signed_string)





import json
import re
import rfc3986
import six
from tempfile import NamedTemporaryFile
from defusedxml.minidom import parseString


def _is_jws(value):
    jws_regex = re.compile(r'^[A-z0-9\-=]+.[A-z0-9\-=]+.[A-z0-9\-_=]+$')
    test_value = value
    if isinstance(value, bytes):
        test_value = value.decode()
    return bool(jws_regex.match(test_value))


def _is_url(value):
    ret = False
    try:
        if (
            (value and isinstance(value, six.string_types))
            and rfc3986.is_valid_uri(value, require_scheme=True)
            and rfc3986.uri_reference(value).scheme.lower() in ['http', 'https']
        ):
            ret = True
    except ValueError:
        pass
    return ret


def bake(image_file, assertion_string, new_file=None):

    svg_doc = parseString(image_file.read())
    image_file.close()

    assertion_node = svg_doc.createElement('openbadges:assertion')
    assertion_node = _populate_assertion_node(assertion_node, assertion_string,
                                              svg_doc)

    svg_body = svg_doc.getElementsByTagName('svg')[0]
    svg_body.setAttribute('xmlns:openbadges', "http://openbadges.org")
    svg_body.insertBefore(assertion_node, svg_body.firstChild)

    if new_file is None:
        new_file = NamedTemporaryFile(suffix='.svg')

    new_file.write(svg_doc.toxml('utf-8'))
    new_file.seek(0)
    return new_file


def _populate_assertion_node(assertion_node, assertion_string, svg_doc):
    assertion = None
    verify_attr = None

    try:
        assertion = json.loads(assertion_string)
    except ValueError:
        pass

    if assertion:
        verify_attr = assertion.get('id')
        if verify_attr is None or not _is_url(verify_attr):  # For hosted badges, use verification URL
            verify_attr = assertion.get('verify', {}).get('url')

    elif _is_jws(assertion_string):  # For signed badges, embed JWS as verify attribute
        verify_attr = assertion_string

    elif _is_url(assertion_string):  # For 0.5 badges, the baking input is the url itself
        verify_attr = assertion_string

    if verify_attr:
        assertion_node.setAttribute('verify', verify_attr)

    character_data = svg_doc.createCDATASection(assertion_string)
    assertion_node.appendChild(character_data)

    return assertion_node


def unbake(image_file):
    svg_doc = parseString(image_file.read())

    assertion_node = svg_doc.getElementsByTagName("openbadges:assertion")[0]
    character_data = None
    verification_data = None
    for node in assertion_node.childNodes:
        if node.nodeType == node.CDATA_SECTION_NODE:
            character_data = node.nodeValue
    try:
        verification_data = assertion_node.attributes['verify'].nodeValue.encode('utf-8')
    except KeyError:
        pass
    return verification_data or character_data





import re

from . import png_bakery, svg_bakery


def check_image_type(image_file):
    if image_file.read(8) == b'\x89PNG\r\n\x1a\n':
        return 'PNG'
    image_file.seek(0)
    # TODO: Use xml library to more accurately detect SVG documents
    try:
        if re.search(b'<svg', image_file.read(256)):
            return 'SVG'
    except TypeError:
        image_file.seek(0)
        if re.search('<svg', image_file.read(256))
            return 'SVG'


def unbake(image_file):
    """
    Return the openbadges content contained in a baked image.
    """
    image_type = check_image_type(image_file)
    image_file.seek(0)
    if image_type == 'PNG':
        return png_bakery.unbake(image_file)
    elif image_type == 'SVG':
        return svg_bakery.unbake(image_file)


def bake(image_file, assertion_json_string, output_file=None):
    """
    Embeds a serialized representation of a badge instance in an image file.
    """
    image_type = check_image_type(image_file)
    image_file.seek(0)
    if image_type == 'PNG':
        return png_bakery.bake(image_file, assertion_json_string, output_file)
    elif image_type == 'SVG':
        return svg_bakery.bake(image_file, assertion_json_string, output_file)





import click

from . import utils


@click.group()
def cli():
    pass


@click.command()
@click.argument('input_file', type=click.File('rb'))
@click.argument('output_file', type=click.File('wb'))
@click.option('--data', type=str, prompt=True)
def bake(input_file, output_file, data):
	"""
	This command bakes Open Badges data into a file and saves
	the result to an output file.

	Positional Arguments:

	\b
	  Input filename:    File must exist.
	\b
	  Output filename:   If file exists, it will be overwritten.
	"""
	output = utils.bake(input_file, data, output_file)

	click.echo(
		"{} is done baking. Remember to let it cool".format(output_file.name)
	)


@click.command()
@click.argument('input_file', type=click.File('rb'))
@click.argument('output_file', type=click.File('wb'), default='-')
def unbake(input_file, output_file):
	"""
	This command extracts Open Badges data from an image and
	prints it to a file or the standard output.

	Positional Arguments:

	\b
	  Input filename:    File must exist.
	\b
	  Output filename:   If file exists, it will be overwritten.
	"""
	click.echo('')
	output_file.write(utils.unbake(input_file).encode('utf8'))
	click.echo('\n')


cli.add_command(bake)
cli.add_command(unbake)

if __name__ == '__main__':
    cli()





import codecs
import hashlib
from io import BytesIO, StringIO
import itertools
import os.path
import png
import re
from tempfile import NamedTemporaryFile


def unbake(imageFile):
    """
    Return the openbadges content contained in a baked PNG file.
    If this doesn't work, return None.

    If there is both an iTXt and tEXt chunk with keyword openbadges,
    the iTXt chunk content will be returned.
    """

    reader = png.Reader(file=imageFile)
    for chunktype, content in reader.chunks():
        if chunktype == b'iTXt' and content.startswith(b'openbadges\x00'):
            return re.sub(b'openbadges[\x00]+', b'', content).decode('utf8')
        elif chunktype == b'tEXt' and content.startswith(b'openbadges\x00'):
            return content.split(b'\x00')[1].decode('utf8')


def bake(imageFile, assertion_string, newfile=None):
    """
    Embeds a serialized representation of a badge instance in a PNG image file.
    """
    encoded_assertion_string = codecs.getwriter('utf-8')(assertion_string)
    reader = png.Reader(file=imageFile)

    if newfile is None:
        newfile = NamedTemporaryFile(suffix='.png')

    chunkheader = b'openbadges\x00\x00\x00\x00\x00'
    chunk_content = chunkheader + encoded_assertion_string.stream.encode('utf-8')
    badge_chunk = (b'iTXt', chunk_content)
    png.write_chunks(newfile, baked_chunks(reader.chunks(), badge_chunk))

    newfile.seek(0)
    return newfile


def baked_chunks(original_chunks, badge_chunk):
    """
    Returns an iterable of chunks that places the Open Badges baked chunk
    and filters out any previous Open Badges chunk that may have existed.
    """
    def is_not_previous_assertion(chunk):
        if chunk[1].startswith(b'openbadges\x00'):
            return False
        return True

    first_slice = next(original_chunks)
    last_slice = list(filter(
        is_not_previous_assertion,
        original_chunks
    ))

    return itertools.chain([first_slice], [badge_chunk], last_slice)



import os
from setuptools import setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='openbadges_bakery',
    version='1.2.5',
    packages=['openbadges_bakery'],
    include_package_data=True,
    license='Apache License 2.0',
    description='A python utility for baking and extracting Open Badges metadata from images.',
    long_description=README,
    url='https://badgr.com',
    author='Concentric Sky',
    author_email='notto@concentricsky.com',
    classifiers=[
        'Environment :: Console',
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Education',
        'Topic :: Utilities',
        'Intended Audience :: Developers'
    ],
    install_requires=[
        'Click>=6.6',
        'pypng>=0.0.20',
        'defusedxml>=0.6.0,<1.0'
    ],
    entry_points="""
        [console_scripts]
        bakery=openbadges_bakery.cli:cli
    """
)
